@@ -357,25 +357,38 @@ ncclResult_t ncclTopoAddNet(struct ncclXmlNode* xmlNet, struct ncclTopoSystem* s
   int dev;
   NCCLCHECK(xmlGetAttrInt(xmlNet, "dev", &dev));
 
+  int64_t netId = NCCL_TOPO_ID(systemId, dev);
   struct ncclTopoNode* net;
-  NCCLCHECK(ncclTopoCreateNode(system, &net, NET, NCCL_TOPO_ID(systemId, dev)));
+  NCCLCHECK(ncclTopoCreateNode(system, &net, NET, netId));
   net->net.dev = dev;
   const char* str;
+  // if not guid is present use the net->id unique id instead, which will be unique within the node/NVLD
   NCCLCHECK(xmlGetAttr(xmlNet, "guid", &str));
-  if (str) sscanf(str, "0x%lx", &net->net.asic);
-  else net->net.asic = dev;
+  net->net.asic = (str) ? strtoull(str, NULL, 16) : netId;
+
 
-  ncclDebugNoWarn = NCCL_GRAPH;
   int mbps;
-  NCCLCHECK(xmlGetAttrIntDefault(xmlNet, "speed", &mbps, 0));
+  NCCLCHECKNOWARN(xmlGetAttrIntDefault(xmlNet, "speed", &mbps, 0), NCCL_GRAPH);
   if (mbps <= 0) mbps = 10000; // Some NICs define speed = -1
   net->net.bw = mbps / 8000.0;
-  if (xmlGetAttrFloat(xmlNet, "latency", &net->net.latency) != ncclSuccess) net->net.latency = 0;
-  NCCLCHECK(xmlGetAttrIntDefault(xmlNet, "port", &net->net.port, 0));
-  NCCLCHECK(xmlGetAttrIntDefault(xmlNet, "gdr", &net->net.gdrSupport, 0));
-  NCCLCHECK(xmlGetAttrIntDefault(xmlNet, "maxconn", &net->net.maxChannels, MAXCHANNELS));
-  NCCLCHECK(xmlGetAttrIntDefault(xmlNet, "coll", &net->net.collSupport, 0));
-  ncclDebugNoWarn = 0;
+  ncclResult_t ret;
+  NOWARN(ret = xmlGetAttrFloat(xmlNet, "latency", &net->net.latency), NCCL_GRAPH);
+  if (ret != ncclSuccess) net->net.latency = 0;
+  NCCLCHECKNOWARN(xmlGetAttrIntDefault(xmlNet, "port", &net->net.port, 0), NCCL_GRAPH);
+  NCCLCHECKNOWARN(xmlGetAttrIntDefault(xmlNet, "gdr", &net->net.gdrSupport, 0), NCCL_GRAPH);
+  NCCLCHECKNOWARN(xmlGetAttrIntDefault(xmlNet, "maxconn", &net->net.maxChannels, MAXCHANNELS), NCCL_GRAPH);
+  NCCLCHECKNOWARN(xmlGetAttrIntDefault(xmlNet, "coll", &net->net.collSupport, 0), NCCL_GRAPH);
+
+  // build the PCI id using the parent PCI link
+  uint64_t hacc[2] = {1, 1};
+  const char* busId = NULL;
+  struct ncclXmlNode* parent = xmlNet->parent;
+  while (parent != NULL && strcmp(parent->name, "pci") != 0) parent = parent->parent;
+  if (parent) NCCLCHECK(xmlGetAttr(parent, "busid", &busId));
+  // If we fail to find the PCIe path, we use the GUID instead.
+  if (busId) eatHash(hacc, busId, strlen(busId));
+  else eatHash(hacc, &net->net.asic);
+  net->net.pciId = digestHash(hacc);
 
   NCCLCHECK(ncclTopoConnectNodes(nic, net, LINK_NET, net->net.bw));
   NCCLCHECK(ncclTopoConnectNodes(net, nic, LINK_NET, net->net.bw));
@@ -998,7 +1011,8 @@ ncclResult_t ncclTopoMakeVnic(struct ncclXml* xml, struct ncclTopoNetInfo* netIn
 
   // Trigger the merge, then get the new device's properties
   int vDevIndex = 0;
-  ncclResult_t ret = netInfo->makeVDevice(&vDevIndex, vProps);
+  ncclResult_t ret;
+  NOWARN(ret = netInfo->makeVDevice(&vDevIndex, vProps), NCCL_GRAPH|NCCL_INIT|NCCL_NET);
   if (ret != ncclSuccess) {
     INFO(NCCL_GRAPH|NCCL_INIT|NCCL_NET, "TOPO/NET : Tried merging multiple devices together and failed. vProps={ndevs=%d, devs=[%d %d %d %d]}. Set NCCL_NET_MERGE_LEVEL=LOC to disable NIC fusion.",
       vProps->ndevs, vProps->devs[0], vProps->devs[1], vProps->devs[2], vProps->devs[3]);
@@ -1582,16 +1596,8 @@ ncclResult_t getLocalNetCountByBw(struct ncclTopoSystem* system, int gpu, int *c
   return ncclSuccess;
 }
 
-enum netDevsPolicy {
-  NETDEVS_POLICY_AUTO = 0x0,
-  NETDEVS_POLICY_ALL = 0x1,
-  NETDEVS_POLICY_MAX = 0x2,
-  NETDEVS_POLICY_UNDEF = 0xffffffff
-};
-
-static enum netDevsPolicy netDevsPolicy = NETDEVS_POLICY_UNDEF;
 static int netDevsPolicyNum = -1;
-
+static enum netDevsPolicy netDevsPolicy = NETDEVS_POLICY_UNDEF;
 static void getNetDevsPolicyOnce() {
   const char* envStr = ncclGetEnv("NCCL_NETDEVS_POLICY");
   if (envStr) {
@@ -1614,6 +1620,18 @@ static void getNetDevsPolicyOnce() {
   if (netDevsPolicy == NETDEVS_POLICY_UNDEF) netDevsPolicy = NETDEVS_POLICY_AUTO;
 }
 
+ncclResult_t ncclTopoGetNetDevsPolicy(enum netDevsPolicy* policy, int* policyNum) {
+  static pthread_once_t onceNetDevsPolicy = PTHREAD_ONCE_INIT;
+  pthread_once(&onceNetDevsPolicy, getNetDevsPolicyOnce);
+  if (netDevsPolicy == NETDEVS_POLICY_MAX && netDevsPolicyNum <= 0) {
+    WARN("Invalid number of network devices = %d for policy MAX", netDevsPolicyNum);
+    return ncclInternalError;
+  }
+  if (policy) *policy = netDevsPolicy;
+  if (policyNum && netDevsPolicyNum >= 0) *policyNum = netDevsPolicyNum;
+  return ncclSuccess;
+}
+
 ncclResult_t ncclTopoGetLocalNet(struct ncclTopoSystem* system, int rank, int channelId, int64_t* id, int* dev) {
   int gpu;
   NCCLCHECK(ncclTopoRankToIndex(system, rank, &gpu, /*showWarn=*/true));
@@ -1626,22 +1644,19 @@ ncclResult_t ncclTopoGetLocalNet(struct ncclTopoSystem* system, int rank, int ch
     return ncclInternalError;
   }
 
-  static pthread_once_t once = PTHREAD_ONCE_INIT;
-  pthread_once(&once,getNetDevsPolicyOnce);
   int netsPerGpu = 0;
-  if (netDevsPolicy == NETDEVS_POLICY_AUTO) {
+  int policyCount = 0;
+  enum netDevsPolicy policy;
+  NCCLCHECK(ncclTopoGetNetDevsPolicy(&policy, &policyCount));
+  if (policy == NETDEVS_POLICY_AUTO) {
     int localGpus[NCCL_TOPO_MAX_NODES];
     int localGpuCount;
     NCCLCHECK(ncclTopoGetLocal(system, NET, localNets[0], GPU, localGpus, &localGpuCount, NULL));
     netsPerGpu = DIVUP(localNetCount, localGpuCount);
-  } else if (netDevsPolicy == NETDEVS_POLICY_ALL) {
+  } else if (policy == NETDEVS_POLICY_ALL) {
     netsPerGpu = localNetCount;
-  } else if (netDevsPolicy == NETDEVS_POLICY_MAX) {
-    if (netDevsPolicyNum <= 0) {
-      WARN("Invalid number of network devices = %d for policy MAX", netDevsPolicyNum);
-      return ncclInternalError;
-    }
-    netsPerGpu = std::min(netDevsPolicyNum, localNetCount);
+  } else if (policy == NETDEVS_POLICY_MAX) {
+    netsPerGpu = std::min(policyCount, localNetCount);
   } else {
     WARN("Unknown netDevs policy");
     return ncclInternalError;
@@ -1655,6 +1670,21 @@ ncclResult_t ncclTopoGetLocalNet(struct ncclTopoSystem* system, int rank, int ch
   return ncclSuccess;
 }
 
+ncclResult_t ncclTopoGetLocalNets(struct ncclTopoSystem* system, int rank, int64_t* localNets, int* localNetCount) {
+  int gpu;
+  NCCLCHECK(ncclTopoRankToIndex(system, rank, &gpu, /*showWarn=*/true));
+  int localNetIndexes[NCCL_TOPO_MAX_NODES];
+  NCCLCHECK(ncclTopoGetLocal(system, GPU, gpu, NET, localNetIndexes, localNetCount, NULL));
+
+  if (*localNetCount == 0) {
+    WARN("Could not find any local path from gpu %d to net.", gpu);
+    return ncclInternalError;
+  }
+  // Convert index to ids
+  for (int n=0; n<*localNetCount; n++) localNets[n] = system->nodes[NET].nodes[localNetIndexes[n]].id;
+  return ncclSuccess;
+}
+
 ncclResult_t ncclTopoGetLocalGpu(struct ncclTopoSystem* system, int64_t netId, int* gpuIndex) {
   ncclResult_t ret = ncclSuccess;
   int netIndex;
